
/*
	view.c

	@description@

	Copyright (C) 1996-1997  Id Software, Inc.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to:

		Free Software Foundation, Inc.
		59 Temple Place - Suite 330
		Boston, MA  02111-1307, USA

	$Id$
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "client.h"
#include "QF/compat.h"
#include "QF/console.h"
#include "draw.h"
#include "host.h"
#include "view.h"
#include "vid.h"

extern byte        gammatable[256];

extern qboolean    V_CheckGamma (void);

extern void        V_CalcIntermissionRefdef (void);
extern void        V_CalcPowerupCshift (void);
extern void        V_CalcRefdef (void);

extern cvar_t     *crosshair;
extern cvar_t     *scr_ofsx;
extern cvar_t     *scr_ofsy;
extern cvar_t     *scr_ofsz;

/*
	V_UpdatePalette
*/
void
V_UpdatePalette (void)
{
	int         i, j;
	qboolean    new;
	byte       *basepal, *newpal;
	byte        pal[768];
	int         r, g, b;
	qboolean    force;

	V_CalcPowerupCshift ();

	new = false;

	for (i = 0; i < NUM_CSHIFTS; i++) {
		if (cl.cshifts[i].percent != cl.prev_cshifts[i].percent) {
			new = true;
			cl.prev_cshifts[i].percent = cl.cshifts[i].percent;
		}
		for (j = 0; j < 3; j++)
			if (cl.cshifts[i].destcolor[j] != cl.prev_cshifts[i].destcolor[j]) {
				new = true;
				cl.prev_cshifts[i].destcolor[j] = cl.cshifts[i].destcolor[j];
			}
	}

// drop the damage value
	cl.cshifts[CSHIFT_DAMAGE].percent -= host_frametime * 150;
	if (cl.cshifts[CSHIFT_DAMAGE].percent <= 0)
		cl.cshifts[CSHIFT_DAMAGE].percent = 0;

// drop the bonus value
	cl.cshifts[CSHIFT_BONUS].percent -= host_frametime * 100;
	if (cl.cshifts[CSHIFT_BONUS].percent <= 0)
		cl.cshifts[CSHIFT_BONUS].percent = 0;

	force = V_CheckGamma ();
	if (!new && !force)
		return;

	basepal = host_basepal;
	newpal = pal;

	for (i = 0; i < 256; i++) {
		r = basepal[0];
		g = basepal[1];
		b = basepal[2];
		basepal += 3;

		for (j = 0; j < NUM_CSHIFTS; j++) {
			r +=
				(cl.cshifts[j].percent * (cl.cshifts[j].destcolor[0] - r)) >> 8;
			g +=
				(cl.cshifts[j].percent * (cl.cshifts[j].destcolor[1] - g)) >> 8;
			b +=
				(cl.cshifts[j].percent * (cl.cshifts[j].destcolor[2] - b)) >> 8;
		}

		newpal[0] = gammatable[r];
		newpal[1] = gammatable[g];
		newpal[2] = gammatable[b];
		newpal += 3;
	}

	VID_ShiftPalette (pal);
}

/*
==================
V_RenderView

The player's clipping box goes from (-16 -16 -24) to (16 16 32) from
the entity origin, so any view position inside that will be valid
==================
*/
extern vrect_t scr_vrect;

void
V_RenderView (void)
{
	if (!cl.worldmodel || cls.signon != SIGNONS)
		return;

// don't allow cheats in multiplayer
	if (cl.maxclients > 1) {
		Cvar_Set (scr_ofsx, "0");
		Cvar_Set (scr_ofsy, "0");
		Cvar_Set (scr_ofsz, "0");
	}

	if (cl.intermission) {				// intermission / finale rendering
		V_CalcIntermissionRefdef ();
	} else {
		if (!cl.paused					/* && (sv.maxclients > 1 || key_dest
										   == key_game) */ )
			V_CalcRefdef ();
	}

	R_PushDlights (vec3_origin);

	R_RenderView ();

	if (crosshair->int_val)
		Draw_Crosshair ();
}

void
BuildGammaTable (float b, float c)
{
	int         i, j;
	int         inf = 0;

	if ((b == 1.0) && (c == 1.0)) {
		for (i = 0; i < 256; i++)
			gammatable[i] = i;
		return;
	}

	for (i = 0; i < 256; i++) {
		if (!(i == 128)) {
			if (i < 128) {
				j = i + (int) ((128 - i) * (1 - c));
			} else {
				j = i + (int) ((i - 128) * (1 - c));
			}
		} else {
			j = i;
		}
		inf = (j * b);					// gamma is brightness now, and
		// positive
		inf = bound (0, inf, 255);
		gammatable[i] = inf;
	}
}
